
/*
 * Arkanoid main module
 *
 * clk 
 *  [input] 25MHz main clock ( 40ns )
 * reset_n
 *  [input] 0 if reset, 1 otherwise
 * x_increment
 *  [input] paddle_x movement
 * left
 *  [input] mouse left button press
 * right
 *  [input] mouse right button press
 * data_ready
 *  [input] 1 if mouse data is ready, 0 otherwise
 * err
 *  [input] 1 if mouse data has an error, 0 otherwise
 * vsync
 *  [input] 0 if vsync event occurs, 1 otherwise
 * read
 *  [output] 1 if mouse data is read already, 0 otherwise
 * paddle_x
 *  [output] current paddle x coordinate 
 * ball_x
 *  [output] current ball x coordinate
 * ball_y
 *  [output] current ball y coordinate
 * life
 *  [output] remaining lives
 *   11 - 2 life
 *   10 - 1 life
 *   01 - 0 life
 *   00 - game over
 * addr
 *  [output] SRAM address 
 * web
 *  [output] 0 if write is enabled, 1 otherwise.
 * data
 *  [inout] SRAM data
 */

`timescale 1ps / 1ps

`define SCREEN_SIZE_X 512
`define SCREEN_SIZE_Y 450

`define BALL_WIDTH 10
`define BALL_HEIGHT 10

`define PADDLE_TOP 375
`define PADDLE_BOTTOM 380
`define PADDLE_WIDTH 50

/*
 * BALL_DELAY
 *  Ball moving delay. The delay time is 2^(BALL_DELAY).
 * MOUSE_SENSITIVITY
 *  The bigger MOUSE_SENSITIVITY, the slower paddle moves.
 */
`define BALL_DELAY 1
`define MOUSE_SENSITIVITY 1

/*
 * BALL_STATE_1
 *  Sets SRAM address for read.
 * BALL_STATE_2
 *  Reads SRAM data.
 * BALL_STATE_3
 *  Changes the direction and delay of the ball.
 * BALL_STATE_4
 *  Sets SRAM address and data for write.
 * BALL_STATE_5
 *  Sets web to 0 for write.
 * BALL_STATE_6
 *  Sets web to 1.
 */
`define BALL_STATE_1 1
`define BALL_STATE_2 2
`define BALL_STATE_3 3
`define BALL_STATE_3_2 7
`define BALL_STATE_4 4
`define BALL_STATE_5 5
`define BALL_STATE_6 6
`define BALL_STATE_7 0

`define RAM_STATE_1 1
`define RAM_STATE_2 2
`define RAM_STATE_3 3
`define RAM_STATE_4 4
`define RAM_STATE_5 5

module main_ark (clk, reset_n, x_increment, left, right, data_ready, err,
  vsyncb, read, paddle_x, ball_x, ball_y, life, addr, web, data, en,
  collision);

  input clk;
  input reset_n;
  input [8:0] x_increment;
  input left;
  input right;
  input data_ready;
  input err;
  input vsyncb;

  output read;
  output [10:0] paddle_x;
  output [10:0] ball_x;
  output [10:0] ball_y;
  output [1:0] life;
  output web;
  output [14:0] addr;
  inout [7:0] data;
  output en;
  output [2:0] collision;

  wire clk;
  wire reset_n;
  wire [8:0] x_increment;
  wire left;
  wire right;
  wire data_ready;
  wire err;
  wire vsyncb;

  reg [2:0] collision;
  reg en;
  reg read;
  reg [10:0] paddle_x;
  reg [10:0] ball_x;
  reg [10:0] ball_y;
  reg [1:0] life;
  reg web;
  reg [14:0] addr;

  tri [7:0] data;
  reg [7:0] data_reg;
  reg oeb;

  assign data = (oeb) ? (data_reg) : (8'bzzzzzzzz);

  wire [10:0] paddle_x_init;
  reg [10:0] paddle_x2;
  wire [10:0] paddle_x3;
  wire [10:0] paddle_x4;
  wire [10:0] paddle_x5;
  wire paddle_x6;
  wire paddle_x7;
  wire [10:0] paddle_x8;
  wire data_ready2;
  reg in_game;
  wire ball_out;
  wire [10:0] ball_x_init;
  wire [10:0] ball_y_init;
  reg [2:0] ball_x_delay;
  wire [2:0] ball_x_delay2;
  reg [2:0] ball_y_delay;
  reg ball_go_x_dir;
  reg ball_go_y_dir;
  reg [`BALL_DELAY:0] ball_delay;
  wire [`BALL_DELAY:0] ball_delay2;
  wire ball_update;
  reg [3:0] block_sel;
  wire [3:0] block_sel2;
  wire [14:0] block_sel_addr;
  reg [8:0] block_data;
  wire valid_block_sel_addr;
  wire [1:0] life2;
  wire left_collision;
  wire right_collision;
  wire top_collision;
  wire paddle_collision;
  wire paddle_left_collision;
  wire paddle_right_collision;
  reg [2:0] ball_x_delay_count;
  reg [2:0] ball_y_delay_count;
  wire [2:0] ball_x_delay_count2;
  wire [2:0] ball_y_delay_count2;
  wire [10:0] ball_x2;
  wire [10:0] ball_y2;
  wire [10:0] ball_xm;
  wire [10:0] ball_xp;
  wire [10:0] ball_ym;
  wire [10:0] ball_yp;
  wire block_collision0;
  wire block_collision1;
  wire block_collision2;
  wire block_collision3;
  wire block_collision4;
  wire block_collision5;
  wire block_collision6;
  wire block_collision7;
  wire block_collision8;
  reg [3:0] ball_state;
  reg need_ball_update;
  wire need_ball_update2;
  reg prev_vsyncb;
  wire ram_init;
  reg [2:0] ram_state;
  wire [14:0] addr2;
  wire [7:0] data2;
  reg [1:0] state_count;
  wire [1:0] state_count2;
  reg [3:0] ball_x_speed;
  wire [3:0] ball_x_speed2;
  reg [3:0] ball_y_speed;
  wire data3;
  reg [7:0] broken_blocks;
  wire [7:0] broken_blocks2;
  wire clear_game;
  wire all_broken;

  assign data_ready2 = ((data_ready) & (~err) & (vsyncb));
  assign paddle_x3 =
    (x_increment[8] ? (paddle_x2 + {2'b11, x_increment} * `MOUSE_SENSITIVITY) :
    (paddle_x2 + x_increment * `MOUSE_SENSITIVITY));
  assign paddle_x4 = (paddle_x3 + (`PADDLE_WIDTH / 2));
  assign paddle_x5 = (paddle_x3 - (`PADDLE_WIDTH / 2));
  assign paddle_x6 = (paddle_x4 > `SCREEN_SIZE_X);
  assign paddle_x7 = (paddle_x5[10]);
  assign paddle_x8 = ((paddle_x7) ? (`PADDLE_WIDTH / 2) :
    ((paddle_x6) ? (`SCREEN_SIZE_X - (`PADDLE_WIDTH / 2)) : paddle_x3));
  assign paddle_x_init = (`SCREEN_SIZE_X / 2);
  assign ball_out = ((ball_y + `BALL_HEIGHT / 2) > (`SCREEN_SIZE_Y - 20));
  assign ball_x_init = ((`SCREEN_SIZE_X / 2) + (`PADDLE_WIDTH / 4));
  assign ball_y_init = (`PADDLE_TOP - (`BALL_HEIGHT / 2));
  function [2:0] calc_ball_x_delay;
    input [10:0] ball_x4;
    input [10:0] paddle_x4;
    begin: calc_ball_x_delay2
      calc_ball_x_delay = 0;
/*      calc_ball_x_delay = (ball_x4 - paddle_x4);
      if (calc_ball_x_delay == 0) begin
        calc_ball_x_delay = 4;
      end
      else begin
        if (calc_ball_x_delay < 0) begin
          calc_ball_x_delay = -calc_ball_x_delay;
        end
        calc_ball_x_delay = ((`PADDLE_WIDTH / 2 - calc_ball_x_delay) >> 3);
      end*/
    end
  endfunction
  assign ball_x_delay2 = calc_ball_x_delay (ball_x, paddle_x);
  assign ball_delay2 = (ball_delay + 1);
//  assign ball_update = ((ball_delay == 0) & (need_ball_update) & (in_game));
  assign ball_update = ((need_ball_update) & (in_game));
  function [14:0] get_block_sel_addr;
    input [10:0] ball_x5;
    input [10:0] ball_y5;
    input [3:0] block_sel;
    begin: get_block_sel_addr2
      reg [4:0] ball_x5_mp;
      reg [4:0] ball_x5_m1;
      reg [4:0] ball_x5_p1;
      reg [4:0] ball_y5_mp;
      reg [4:0] ball_y5_m1;
      reg [4:0] ball_y5_p1;
      ball_x5_mp = (ball_x5[9:5]);
      ball_x5_m1 = (ball_x5_mp - 1);
      ball_x5_p1 = (ball_x5_mp + 1);
      ball_y5_mp = (ball_y5[8:4]);
      ball_y5_m1 = (ball_y5_mp - 1);
      ball_y5_p1 = (ball_y5_mp + 1);
      get_block_sel_addr = 15'b111111111111111;
      case (block_sel)
        0: begin
          if (!((ball_x5_mp == 0) || (ball_y5_mp == 0) ||
            (ball_y5_m1[4:3] != 0))) begin
            get_block_sel_addr = {7'b0000000, ball_y5_m1[3:0],
              ball_x5_m1[3:0]};
          end
        end
        1: begin
          if (!((ball_y5_mp == 0) || (ball_y5_m1[4:3] != 0))) begin
            get_block_sel_addr = {7'b0000000, ball_y5_m1[3:0],
              ball_x5_mp[3:0]};
          end
        end
        2: begin
          if (!((ball_x5_mp == 5'b01111) || (ball_y5_mp == 0) ||
            (ball_y5_m1[4:3] != 0))) begin
            get_block_sel_addr = {7'b0000000, ball_y5_m1[3:0],
              ball_x5_p1[3:0]};
          end
        end
        3: begin
          if (!((ball_x5_mp == 0) || (ball_y5_mp[4:3] != 0))) begin
            get_block_sel_addr = {7'b0000000, ball_y5_mp[3:0],
              ball_x5_m1[3:0]};
          end
        end
        4: begin
          get_block_sel_addr = {7'b0000000, ball_y5_mp[3:0], ball_x5_mp[3:0]};
        end
        5: begin
          if (!((ball_x5_mp == 5'b01111) || (ball_y5_mp[4:3] != 0))) begin
            get_block_sel_addr = {7'b0000000, ball_y5_mp[3:0],
              ball_x5_p1[3:0]};
          end
        end
        6: begin
          if (!((ball_x5_mp == 0) || (ball_y5_mp == 5'b01111) ||
            (ball_y5_p1[4:3] != 0))) begin
            get_block_sel_addr = {7'b0000000, ball_y5_p1[3:0],
              ball_x5_m1[3:0]};
          end
        end
        7: begin
          if (!((ball_y5_mp == 5'b01111) || (ball_y5_p1[4:3] != 0))) begin
            get_block_sel_addr = {7'b0000000, ball_y5_p1[3:0],
              ball_x5_mp[3:0]};
          end
        end
        8: begin
          if (!((ball_x5_mp == 5'b01111) || (ball_y5_mp == 5'b01111) ||
            (ball_y5_p1[4:3] != 0))) begin
            get_block_sel_addr = {7'b0000000, ball_y5_p1[3:0],
              ball_x5_p1[3:0]};
          end
        end
        default: begin
        end
      endcase
    end
  endfunction
  assign block_sel_addr = get_block_sel_addr (ball_x, ball_y, block_sel);
  assign block_sel2 = ((block_sel >= 8) ? (0) : (block_sel + 1));
  assign valid_block_sel_addr = (block_sel_addr != 15'b111111111111111);
  assign life2 = (life - 1);
  assign left_collision = ((ball_x) <= (`BALL_WIDTH / 2)) && (~ball_go_x_dir);
  assign right_collision = ((ball_x) >= (`SCREEN_SIZE_X - (`BALL_WIDTH / 2)))
    && (ball_go_x_dir);
  assign top_collision = ((ball_y) <= (`BALL_WIDTH / 2)) && (~ball_go_y_dir);
  assign paddle_collision = ((ball_y + `BALL_HEIGHT / 2) <= (`PADDLE_BOTTOM))
    && ((ball_y + `BALL_HEIGHT / 2) >= (`PADDLE_TOP))
    && ((ball_x) <= (paddle_x + `PADDLE_WIDTH / 2))
    && ((ball_x) >= (paddle_x - `PADDLE_WIDTH / 2)) && (ball_go_y_dir);
  assign paddle_left_collision = ((ball_x) >= (paddle_x - `PADDLE_WIDTH / 2))
    && (ball_x <= paddle_x) && (paddle_collision);
  assign paddle_right_collision = ((ball_x) <= (paddle_x + `PADDLE_WIDTH / 2))
    && (ball_x >= paddle_x) && (paddle_collision);
  assign ball_x_delay_count2 = ((ball_x_delay == ball_x_delay_count) ? (0) :
    (ball_x_delay_count + 1));
  assign ball_y_delay_count2 = ((ball_y_delay == ball_y_delay_count) ? (0) :
    (ball_y_delay_count + 1));
  assign ball_x2 = ((ball_go_x_dir) ? (ball_x + ball_x_speed) :
    (ball_x - ball_x_speed));
  assign ball_y2 = ((ball_go_y_dir) ? (ball_y + ball_y_speed) :
    (ball_y - ball_y_speed));
  assign ball_xm = (ball_x - (`BALL_WIDTH / 2 + 1));
  assign ball_xp = (ball_x + (`BALL_WIDTH / 2 + 1));
  assign ball_ym = (ball_y - (`BALL_HEIGHT / 2 + 1));
  assign ball_yp = (ball_y + (`BALL_HEIGHT / 2 + 1));
  assign block_collision0 = (block_data[8]) &&
    (~ball_go_x_dir) && (~ball_go_y_dir) &&
    (ball_xm[5] != ball_x[5]) &&
    (ball_ym[4] != ball_y[4]);
  assign block_collision1 = (block_data[7]) && (~ball_go_y_dir) &&
    (ball_ym[4] != ball_y[4]);
  assign block_collision2 = (block_data[6]) &&
    (ball_go_x_dir) && (~ball_go_y_dir) &&
    (ball_xp[5] != ball_x[5]) &&
    (ball_ym[4] != ball_y[4]);
  assign block_collision3 = (block_data[5]) && (~ball_go_x_dir) &&
    (ball_xm[5] != ball_x[5]);
  assign block_collision4 = 0;
  assign block_collision5 = (block_data[3]) && (ball_go_x_dir) &&
    (ball_xp[5] != ball_x[5]);
  assign block_collision6 = (block_data[2]) &&
    (~ball_go_x_dir) && (ball_go_y_dir) &&
    (ball_xm[5] != ball_x[5]) &&
    (ball_yp[4] != ball_y[4]);
  assign block_collision7 = (block_data[1]) && (ball_go_y_dir) &&
    (ball_yp[4] != ball_y[4]);
  assign block_collision8 = (block_data[0]) &&
    (ball_go_x_dir) && (ball_go_y_dir) &&
    (ball_xp[5] != ball_x[5]) &&
    (ball_yp[4] != ball_y[4]);
  assign need_ball_update2 = ((prev_vsyncb) & (~vsyncb));
  assign ram_init = (~in_game) & (life == 3) & (ram_state != `RAM_STATE_5);
  assign addr2 = (addr + 1);
  assign data2 = ((addr2 < 15'b000000010000000) ? ((addr2 % 7) + 1) : (0));
  assign state_count2 = (state_count + 1);
  function [2:0] calc_ball_x_speed;
    input [10:0] ball_x4;
    input [10:0] paddle_x4;
    reg [10:0] speed;
    begin: calc_ball_x_speed2
      speed = (ball_x4 - paddle_x4);
      if (speed == 0) begin
        calc_ball_x_speed = 1;
      end
      else begin
        if (speed[10] == 1) begin
          speed = -speed;
        end
        speed = (speed >> 3) + 2;
        calc_ball_x_speed = speed[2:0];
      end
    end
  endfunction
  assign ball_x_speed2 = calc_ball_x_speed(ball_x, paddle_x);
  assign data3 = ((data) ? (1) : (0));
  assign broken_blocks2 = (broken_blocks + 1);
  assign all_broken = (broken_blocks == 128);
  assign clear_game = (!((life == 3) && (in_game == 0)));

  /* ball delay count */
  always
    @(negedge vsyncb or negedge reset_n) begin
      if (~reset_n) begin
        ball_delay <= 0;
      end
      else begin
        ball_delay <= ball_delay2;
      end
    end

  /* paddle control */
  always
    @(posedge clk or negedge reset_n) begin
      if (~reset_n) begin
        paddle_x <= paddle_x_init;
        paddle_x2 <= paddle_x_init;
        read <= 0;
        in_game <= 0;
        life <= 3;
      end
      else begin
        if (read) begin
          read <= 0;
        end
        else if (data_ready2) begin
          if (in_game) begin
            paddle_x2 <= paddle_x8;
          end
          if (left) begin
            in_game <= 1;
          end
          read <= 1;
        end
        if (ball_out) begin
          paddle_x2 <= paddle_x_init;
          in_game <= 0;
          life <= life2;
        end
        if (~vsyncb) begin
          paddle_x <= paddle_x2;
        end
      end
    end

  /* ball control */
  always
    @(posedge clk or negedge reset_n) begin
      if (~reset_n) begin
        ball_x <= ball_x_init;
        ball_y <= ball_y_init;
        ball_go_x_dir <= 1;
        ball_go_y_dir <= 0;
        ball_x_delay <= 0;
        ball_y_delay <= 0;
        block_sel <= 0;
        ball_state <= `BALL_STATE_1;
        ball_x_delay_count <= 0;
        ball_y_delay_count <= 0;
        need_ball_update <= 0;
        data_reg <= 0;
        web <= 1;
        oeb <= 0;
        addr <= 15'b111111111111111;
        ram_state <= `RAM_STATE_1;
        state_count <= 0;
        en <= 0;
        ball_x_speed <= 4;
        ball_y_speed <= 4;
        collision <= 0;
      end
      else begin
        if (~vsyncb) begin
          if (ram_init) begin
            en <= 0;
            broken_blocks <= 0;
            if (!state_count) begin
              case (ram_state)
                `RAM_STATE_1: begin
                  addr <= addr2;
                  data_reg <= data2;
                  oeb <= 1;
                  ram_state <= `RAM_STATE_2;
                end
                `RAM_STATE_2: begin
                  web <= 0;
                  ram_state <= `RAM_STATE_3;
                end
                `RAM_STATE_3: begin
                  web <= 1;
                  ram_state <= `RAM_STATE_4;
                end
                `RAM_STATE_4: begin
                  oeb <= 0 ;
                  if (addr2) begin
                    ram_state <= `RAM_STATE_1;
                  end
                  else begin
                    ram_state <= `RAM_STATE_5;
                    en <= 1;
                  end
                end
                `RAM_STATE_5: begin
                end
                default: begin
                end
              endcase
            end
            state_count <= state_count2;
          end
        end
        if (need_ball_update2) begin
          need_ball_update <= 1;
        end
        prev_vsyncb <= vsyncb;
        collision <= 0;
        if (all_broken) begin
          if (clear_game) begin
            ball_y <= `SCREEN_SIZE_Y;
          end
        end
        if (ball_out) begin
          ball_x <= ball_x_init;
          ball_y <= ball_y_init;
          ball_go_x_dir <= 1;
          ball_go_y_dir <= 0;
          ball_x_delay <= 0;
          ball_y_delay <= 0;
          block_sel <= 0;
          ball_state <= `BALL_STATE_1;
          ball_x_delay_count <= 0;
          ball_y_delay_count <= 0;
          need_ball_update <= 0;
          data_reg <= 0;
          web <= 1;
          oeb <= 0;
          addr <= 15'b111111111111111;
          ram_state <= `RAM_STATE_1;
          state_count <= 0;
          ball_x_speed <= 4;
          ball_y_speed <= 4;
          collision <= 3'b001;
        end
        else if (ball_update) begin
          if (!state_count) begin
            case (ball_state)
              `BALL_STATE_1: begin
                if (valid_block_sel_addr) begin
                  addr <= block_sel_addr;
                end
                ball_state <= `BALL_STATE_2;
              end
              `BALL_STATE_2: begin
                if (valid_block_sel_addr) begin
                  block_data <= {block_data[7:0], data3};
                end
                else begin
                  block_data <= {block_data[7:0], 1'b0};
                end
                if (block_sel2) begin
                  ball_state <= `BALL_STATE_1;
                end
                else begin
                  ball_state <= `BALL_STATE_3;
                end
                block_sel <= block_sel2;
              end
              `BALL_STATE_3: begin
                collision <= 0;
                if (left_collision) begin
                  ball_go_x_dir <= 1;
                  collision <= 3'b101;
                end
                else if (right_collision) begin
                  ball_go_x_dir <= 0;
                  collision <= 3'b101;
                end
                else if (top_collision) begin
                  ball_go_y_dir <= 1;
                  collision <= 3'b101;
                end
                else if (paddle_collision) begin
                  if (paddle_left_collision) begin
                    ball_go_x_dir <= 0;
                  end
                  if (paddle_right_collision) begin
                    ball_go_x_dir <= 1;
                  end
                  ball_x_delay <= ball_x_delay2;
                  ball_x_speed <= ball_x_speed2;
                  ball_go_y_dir <= 0;
                  collision <= 3'b011;
                end
                else if (block_collision1) begin
                  block_data[7] <= 0;
                  ball_go_y_dir <= 1;
                  collision <= 3'b111;
                  broken_blocks <= broken_blocks2;
                end
                else if (block_collision3) begin
                  block_data[5] <= 0;
                  ball_go_x_dir <= 1;
                  collision <= 3'b111;
                  broken_blocks <= broken_blocks2;
                end
                else if (block_collision5) begin
                  block_data[3] <= 0;
                  ball_go_x_dir <= 0;
                  collision <= 3'b111;
                  broken_blocks <= broken_blocks2;
                end
                else if (block_collision7) begin
                  block_data[1] <= 0;
                  ball_go_y_dir <= 0;
                  collision <= 3'b111;
                  broken_blocks <= broken_blocks2;
                end
                else if (block_collision0) begin
                  block_data[8] <= 0;
                  ball_go_y_dir <= 1;
                  collision <= 3'b111;
                  broken_blocks <= broken_blocks2;
                end
                else if (block_collision2) begin
                  block_data[6] <= 0;
                  ball_go_y_dir <= 1;
                  collision <= 3'b111;
                  broken_blocks <= broken_blocks2;
                end
                else if (block_collision6) begin
                  block_data[2] <= 0;
                  ball_go_x_dir <= 1;
                  collision <= 3'b111;
                  broken_blocks <= broken_blocks2;
                end
                else if (block_collision8) begin
                  block_data[0] <= 0;
                  ball_go_x_dir <= 0;
                  collision <= 3'b111;
                  broken_blocks <= broken_blocks2;
                end
                else begin
                  ball_state <= `BALL_STATE_4;
                end
              end
              `BALL_STATE_3_2: begin
                if (!ball_x_delay_count) begin
                  ball_x <= ball_x2;
                end
                if (!ball_y_delay_count) begin
                  ball_y <= ball_y2;
                end
                ball_x_delay_count <= ball_x_delay_count2;
                ball_y_delay_count <= ball_y_delay_count2;
                need_ball_update <= 0;
                ball_state <= `BALL_STATE_1;
              end
              `BALL_STATE_4: begin
                if (valid_block_sel_addr) begin
                  addr <= block_sel_addr;
                  data_reg <= {7'b0000000, block_data[8]};
                  oeb <= 1;
                end
                block_data <= {block_data[7:0], 1'b0};
                ball_state <= `BALL_STATE_5;
                collision <= 0;
              end
              `BALL_STATE_5: begin
                if (valid_block_sel_addr) begin
                  if (!data_reg) begin
                    web <= 0;
                  end
                end
                ball_state <= `BALL_STATE_6;
              end
              `BALL_STATE_6: begin
                if (valid_block_sel_addr) begin
                  web <= 1;
                end
                ball_state <= `BALL_STATE_7;
              end
              `BALL_STATE_7: begin
                if (valid_block_sel_addr) begin
                  oeb <= 0;
                end
                if (block_sel2) begin
                  ball_state <= `BALL_STATE_4;
                end
                else begin
                  ball_state <= `BALL_STATE_3_2;
                end
                block_sel <= block_sel2;
              end
            endcase
          end
          state_count <= state_count2;
        end
      end
    end
endmodule
